# Copyright 2022 Soul Machines Ltd

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from typing import List

from smskillsdk.models.api import Memory, MemoryScope
from smskillsdk.utils import memory


class TestMemoryUtils(unittest.TestCase):
    def test_serialize(self):
        memory_dict = {
            "key1": "value1",
            "key2": True,
            "key3": None,
        }

        memory_array = memory.serialize_memory(memory_dict)
        self.assertEqual(len(memory_array), 3)

        for memory_element in memory_array:
            self.assertIsNone(memory_element.session_id)
            self.assertIn(memory_element.name, memory_dict)
            self.assertEqual(memory_element.value, memory_dict[memory_element.name])

    def test_serialize_with_session_id(self):
        memory_dict = {
            "key1": "value1",
            "key2": True,
            "key3": None,
        }
        session_id = "123"

        memory_array = memory.serialize_memory(memory_dict, session_id)
        self.assertEqual(len(memory_array), 3)

        for memory_element in memory_array:
            self.assertIn(memory_element.name, memory_dict)
            self.assertEqual(memory_element.value, memory_dict[memory_element.name])
            self.assertEqual(memory_element.session_id, session_id)

    def test_serialize_with_scope(self):
        memory_dict = {
            "key1": "value1",
            "key2": "value2",
            "key3": "value3",
        }
        scope = MemoryScope.PUBLIC

        memory_array = memory.serialize_memory(memory_dict, scope=scope)
        self.assertEqual(len(memory_array), 3)

        for memory_element in memory_array:
            self.assertIn(memory_element.name, memory_dict)
            self.assertEqual(memory_element.value, memory_dict[memory_element.name])
            self.assertEqual(memory_element.scope, scope)

    def test_deserialize(self):
        memory_array = [
            Memory(**{"name": "key1", "value": "value1"}),
            Memory(**{"name": "key2", "value": 1}),
            Memory(**{"name": "key3", "value": {"subkey2": "subvalue2"}}),
        ]

        memory_dict = memory.deserialize_memory(memory_array)
        self.assertEqual(len(memory_dict), 3)

        for memory_element in memory_array:
            self.assertIn(memory_element.name, memory_dict)
            self.assertEqual(memory_element.value, memory_dict[memory_element.name])

    def test_deserialize_with_session_id(self):
        memory_array = [
            Memory(**{"name": "key1", "value": "value1", "session_id": "123"}),
            Memory(**{"name": "key2", "value": 1}),
            Memory(
                **{
                    "name": "key3",
                    "value": {"subkey2": "subvalue2"},
                    "session_id": "123",
                }
            ),
        ]

        memory_dict = memory.deserialize_memory(memory_array, "123")
        self.assertEqual(set(memory_dict.keys()), {"key1", "key3"})

    def test_deserialize_with_scope(self):
        memory_array = [
            Memory(**{"name": "key1", "value": "1", "scope": "PUBLIC"}),
            Memory(**{"name": "key2", "value": "2", "scope": "PRIVATE"}),
            Memory(**{"name": "key3", "value": "3"}),
            Memory(**{"name": "key4", "value": "4", "scope": "PRIVATE"}),
            Memory(**{"name": "key5", "value": "5", "scope": "PUBLIC"}),
        ]

        memory_dict = memory.deserialize_memory(memory_array, scope=MemoryScope.PUBLIC)
        self.assertEqual(set(memory_dict.keys()), {"key1", "key5"})

        memory_dict = memory.deserialize_memory(memory_array, scope=MemoryScope.PRIVATE)
        self.assertEqual(set(memory_dict.keys()), {"key2", "key3", "key4"})

        memory_dict = memory.deserialize_memory(memory_array)
        self.assertEqual(
            set(memory_dict.keys()), {"key1", "key2", "key3", "key4", "key5"}
        )

    def test_set_memory_value(self):
        memories: List[Memory] = list()

        memory.set_memory_value(memories, "key1", "value1")
        self.assertEqual(len(memories), 1)
        self.assertEqual(memories[0].name, "key1")
        self.assertEqual(memories[0].value, "value1")

        memory.set_memory_value(memories, "key1", "value2", "sessionid1")
        self.assertEqual(len(memories), 2)
        self.assertEqual(memories[0].name, "key1")
        self.assertEqual(memories[0].value, "value1")
        self.assertIsNone(memories[0].session_id)
        self.assertEqual(memories[-1].name, "key1")
        self.assertEqual(memories[-1].value, "value2")
        self.assertEqual(memories[-1].session_id, "sessionid1")

        memory.set_memory_value(memories, "key1", "value3")
        self.assertEqual(len(memories), 2)
        self.assertEqual(memories[0].name, "key1")
        self.assertEqual(memories[0].value, "value3")
        self.assertEqual(memories[-1].name, "key1")
        self.assertEqual(memories[-1].value, "value2")

        memory.set_memory_value(memories, "key1", "value4", "sessionid1")
        self.assertEqual(len(memories), 2)
        self.assertEqual(memories[0].name, "key1")
        self.assertEqual(memories[0].value, "value3")
        self.assertEqual(memories[-1].name, "key1")
        self.assertEqual(memories[-1].value, "value4")

        memory.set_memory_value(memories, "key1", "value5", scope=MemoryScope.PRIVATE)
        self.assertEqual(len(memories), 2)
        self.assertEqual(memories[0].name, "key1")
        self.assertEqual(memories[0].value, "value5")
        self.assertEqual(memories[-1].name, "key1")
        self.assertEqual(memories[-1].value, "value4")

        memory.set_memory_value(memories, "key2", "value6")
        self.assertEqual(len(memories), 3)
        self.assertEqual(memories[-1].name, "key2")
        self.assertEqual(memories[-1].value, "value6")

        memory.set_memory_value(memories, "key2", "value7", scope=MemoryScope.PUBLIC)
        self.assertEqual(len(memories), 4)
        self.assertEqual(memories[-1].name, "key2")
        self.assertEqual(memories[-1].value, "value7")

    def test_get_memory_value(self):
        memories: List[Memory] = list()

        self.assertEqual(memory.get_memory_value(memories, "key1")[0], False)

        memories.append(Memory(**{"name": "key1", "value": "value1"}))
        self.assertEqual(memory.get_memory_value(memories, "key1"), (True, "value1"))
        self.assertFalse(
            memory.get_memory_value(memories, "key1", scope=MemoryScope.PUBLIC)[0]
        )
        self.assertTrue(
            memory.get_memory_value(memories, "key1", scope=MemoryScope.PRIVATE)[0]
        )

        memories.append(
            Memory(**{"name": "key1", "value": "value2", "session_id": "sessionid1"})
        )
        self.assertEqual(memory.get_memory_value(memories, "key1"), (True, "value1"))
        self.assertEqual(
            memory.get_memory_value(memories, "key1", "sessionid1"), (True, "value2")
        )
        self.assertFalse(memory.get_memory_value(memories, "key1", "sessionid2")[0])

        memories.append(
            Memory(**{"name": "key1", "value": "value3", "scope": "PUBLIC"})
        )
        self.assertEqual(memory.get_memory_value(memories, "key1"), (True, "value1"))
        self.assertEqual(
            memory.get_memory_value(memories, "key1", scope=MemoryScope.PRIVATE),
            (True, "value1"),
        )
        self.assertEqual(
            memory.get_memory_value(memories, "key1", scope=MemoryScope.PUBLIC),
            (True, "value3"),
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 Soul Machines Ltd

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# generated by datamodel-codegen:
#   filename:  skill-api.yml
#   timestamp: 2022-03-30T02:12:35+00:00

from __future__ import annotations

from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Extra, Field


class Intent(BaseModel):
    name: Optional[str] = Field(
        None,
        description="""Name of intent that handled the request or 'NO_MATCH' if a
                       fallback is hit""",
    )
    confidence: Optional[float] = Field(
        None, description="Confidence of the intent matched in the range 0-1"
    )


class Variables(BaseModel):
    class Config:
        extra = Extra.allow

    public: Optional[Dict[str, Any]] = None


class Output(BaseModel):
    text: Optional[str] = None
    variables: Optional[Variables] = None


class HistoryInput(BaseModel):
    text: Optional[str] = Field(None, description="Input text")


class ProcessingStage(Enum):
    pre = "pre"
    post = "post"


class MemoryScope(Enum):
    PRIVATE = "PRIVATE"
    PUBLIC = "PUBLIC"


class Memory(BaseModel):
    name: str = Field(..., description="Identifier for the memory value")
    value: Any = Field(..., description="Memory value (can be of any type)")
    session_id: Optional[str] = Field(
        None, description="Optional session ID to identify this memory value"
    )
    scope: Optional[MemoryScope] = None


class ExecuteResponse(BaseModel):
    output: Optional[Output] = None
    intent: Optional[Intent] = None
    memory: Optional[List[Memory]] = None
    endConversation: bool = Field(
        ...,
        description="""Used to indicate whether the current interaction with the skill
                       should be concluded.  Set to false if you want the skill to
                       handle further conversation turns with the user.\n""",
    )


class SessionRequest(BaseModel):
    projectId: str = Field(
        ..., description="(Future functionality) Salted hash of DDNA Studio Project ID"
    )
    sessionId: str
    config: Optional[Dict[str, Any]] = None
    memory: Optional[List[Memory]] = None


class SessionResponse(BaseModel):
    memory: Optional[List[Memory]] = None


class HistoryOutputMetadata(BaseModel):
    endConversation: Optional[bool] = Field(
        None,
        description="""Used to indicate whether the current interaction with the skill
                       should be concluded.  Set to false if you want the skill to
                       handle further conversation turns with the user.\n""",
    )
    memory: Optional[List[Memory]] = Field(
        None,
        description="""Information stored from the most recent response received from
        the endpoint""",
    )
    projectId: Optional[str] = Field(
        None, description="(Future functionality) Salted hash of DDNA Studio Project ID"
    )
    sessionId: Optional[str] = Field(
        None, description="A unique identifier for this particular session"
    )


class HistoryOutput(BaseModel):
    text: Optional[str] = Field(None, description="Input text")
    variables: Optional[Variables] = None
    metadata: Optional[HistoryOutputMetadata] = None


class HistoryItem(BaseModel):
    input: Optional[HistoryInput] = None
    output: Optional[HistoryOutput] = None


class ConversationHistory(BaseModel):
    __root__: List[HistoryItem]


class Context(BaseModel):
    class Config:
        extra = Extra.allow

    Processing_Stage: Optional[ProcessingStage] = None
    Conversation_History: Optional[ConversationHistory] = None


class ExecuteRequest(BaseModel):
    projectId: str = Field(
        ..., description="(Future functionality) Salted hash of DDNA Studio Project ID"
    )
    sessionId: str = Field(
        ..., description="A unique identifier for this particular session"
    )
    intent: Optional[Intent] = None
    text: str = Field(..., description="User input")
    memory: List[Memory] = Field(
        ...,
        description="""Information stored from the most recent response received from
        the endpoint""",
    )
    context: Optional[Context] = Field(
        None,
        description="""Contains the context of the request, is also used to send DP
                       specific data e.g. EQ data. Might contain Processing_Stage
                       stating if a request is for a pre or post stage. Might contain
                       Conversation_History contains history to up to 5 conversation
                       turns.\n""",
    )
    config: Optional[Dict[str, Any]] = Field(
        None,
        description="""Skill configuration data will be included in this object. This
        property will only be included if the skill is stateless (i.e. no
        sessionEndpoint is provided).\n""",
    )



# Copyright 2022 Soul Machines Ltd

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for working with memory arrays"""
from typing import Any, Dict, List, Tuple, Union

from smskillsdk.models.api import Memory, MemoryScope


def _has_valid_session_id_and_scope(
    memory: Memory,
    session_id: Union[str, None],
    scope: Union[MemoryScope, None],
    strict_session_id_match: bool = False,
) -> bool:
    if (
        strict_session_id_match or session_id is not None
    ) and memory.session_id != session_id:
        return False

    if (
        scope == MemoryScope.PUBLIC and memory.scope in [None, MemoryScope.PRIVATE]
    ) or (scope == MemoryScope.PRIVATE and memory.scope == MemoryScope.PUBLIC):
        return False

    return True


def serialize_memory(
    data: dict,
    session_id: Union[str, None] = None,
    scope: MemoryScope = MemoryScope.PRIVATE,
) -> List[Memory]:
    """Convert a Python dict into a list of Memory objects

    Key-value pairs will be converted into objects with these properties

    {
        "name": <key: Any>,
        "value": <value: Any>
        "session_id": <session_id: str> (optional)
        "scope": <PUBLIC | PRIVATE (default)>
    }
    """
    return [
        Memory(name=key, value=value, session_id=session_id, scope=scope)
        for key, value in data.items()
    ]


def deserialize_memory(
    memories: List[Memory],
    session_id: Union[str, None] = None,
    scope: Union[MemoryScope, None] = None,
) -> Dict[str, Any]:
    """Convert a list of memory objects to a dictionary

    It is assumed memory objects will have name/value attributes. If multiple memory
    objects share the same name values, the latest value in the list will be used. If
    a session_id is provided, this will be checked and used to filter out memories
    without a matching session_id.
    """
    return {
        memory.name: memory.value
        for memory in memories
        if _has_valid_session_id_and_scope(memory, session_id, scope)
    }


def set_memory_value(
    memories: List[Memory],
    key: str,
    value: Any,
    session_id: Union[str, None] = None,
    scope: MemoryScope = MemoryScope.PRIVATE,
) -> None:
    """Sets a value in the memory array corresponding to an optional session_id/scope

    Modifies the memories array in-place. A session ID can be provided to only modify
    a memory value with a matching session ID.
    """
    found = False

    for memory in memories:
        if not _has_valid_session_id_and_scope(memory, session_id, scope, True):
            continue

        if memory.name == key:
            memory.value = value
            found = True

    if not found:
        memories.append(
            Memory(name=key, value=value, session_id=session_id, scope=scope)
        )


def get_memory_value(
    memories: List[Memory],
    key: str,
    session_id: Union[str, None] = None,
    scope: Union[MemoryScope, None] = None,
) -> Tuple[bool, Any]:
    """Gets a value from a memory array corresponding to a session_id/scope"""
    for memory in memories:
        if not _has_valid_session_id_and_scope(memory, session_id, scope):
            continue

        if memory.name == key:
            return True, memory.value

    return False, None
